@@ -15,6 +15,7 @@
 # limitations under the License.
 
 import mesonbuild
+from .. import build
 import sys, os, subprocess, time, datetime, pickle, multiprocessing, json
 import concurrent.futures as conc
 import argparse
@@ -44,13 +45,15 @@ parser.add_argument('args', nargs='+')
 
 
 class TestRun():
-    def __init__(self, res, returncode, should_fail, duration, stdo, stde, cmd):
+    def __init__(self, res, returncode, should_fail, duration, stdo, stde, cmd,
+                 env):
         self.res = res
         self.returncode = returncode
         self.duration = duration
         self.stdo = stdo
         self.stde = stde
         self.cmd = cmd
+        self.env = env
         self.should_fail = should_fail
 
     def get_log(self):
@@ -58,7 +61,9 @@ class TestRun():
         if self.cmd is None:
             res += 'NONE\n'
         else:
-            res += ' '.join(self.cmd) + '\n'
+            res += "\n%s %s\n" %(' '.join(
+                ["%s='%s'" % (k, v) for k, v in self.env.items()]),
+                ' ' .join(self.cmd))
         if self.stdo:
             res += '--- stdout ---\n'
             res += self.stdo
@@ -84,7 +89,8 @@ def write_json_log(jsonlogfile, test_name, result):
               'result' : result.res,
               'duration' : result.duration,
               'returncode' : result.returncode,
-              'command' : result.cmd}
+              'command' : result.cmd,
+              'env' : result.env}
     if result.stde:
         jresult['stderr'] = result.stde
     jsonlogfile.write(json.dumps(jresult) + '\n')
@@ -123,6 +129,9 @@ def run_single_test(wrap, test):
             cmd = wrap + cmd + test.cmd_args
         starttime = time.time()
         child_env = os.environ.copy()
+        if isinstance(test.env, build.EnvironmentVariables):
+            test.env = test.env.get_env(child_env)
+
         child_env.update(test.env)
         if len(test.extra_paths) > 0:
             child_env['PATH'] = child_env['PATH'] + ';'.join([''] + test.extra_paths)
@@ -162,7 +171,7 @@ def run_single_test(wrap, test):
         else:
             res = 'FAIL'
         returncode = p.returncode
-    return TestRun(res, returncode, test.should_fail, duration, stdo, stde, cmd)
+    return TestRun(res, returncode, test.should_fail, duration, stdo, stde, cmd, test.env)
 
 def print_stats(numlen, tests, name, result, i, logfile, jsonlogfile):
     global collected_logs, error_count, options
